/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "faradayConcentrationFvPatchScalarField.H"
#include "fvPatchFieldMapper.H"
#include "volFields.H"
#include "surfaceFields.H"
#include "addToRunTimeSelectionTable.H"

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::faradayConcentrationFvPatchScalarField::faradayConcentrationFvPatchScalarField
(
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF
)
:
    fixedGradientFvPatchScalarField(p, iF),
    molarMass_(1.0)
{}

// this constructor is called in the solver and decomposePar
Foam::faradayConcentrationFvPatchScalarField::faradayConcentrationFvPatchScalarField
(
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF,
    const dictionary& dict
)
:
    fixedGradientFvPatchScalarField(p, iF),
    molarMass_(dict.lookupOrDefault<scalar>("molarMass", 1.0))
{
    if (dict.found("value"))                    // initial value
    {
        fvPatchField<scalar>::operator=
        (
            scalarField("value", dict, p.size())
        );
    }
    else
    {
        fvPatchField<scalar>::operator=(patchInternalField());
    }
}

// probably only needed for mapFields
Foam::faradayConcentrationFvPatchScalarField::faradayConcentrationFvPatchScalarField
(
    const faradayConcentrationFvPatchScalarField& ptf,
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF,
    const fvPatchFieldMapper& mapper
)
:
    fixedGradientFvPatchScalarField(p, iF),
    molarMass_(ptf.molarMass_)
{
    patchType() = ptf.patchType();

    // Map gradient. Set unmapped values and overwrite with mapped ptf
    gradient() = 0.0;
    gradient().map(ptf.gradient(), mapper);

    // Evaluate the value field from the gradient if the internal field is valid
    if (notNull(iF) && iF.size())
    {
        scalarField::operator=
        (
            //patchInternalField() + gradient()/patch().deltaCoeffs()
            // ***HGW Hack to avoid the construction of mesh.deltaCoeffs
            // which fails for AMI patches for some mapping operations
            patchInternalField() + gradient()*(patch().nf() & patch().delta())
        );
    }
    else
    {
        // Enforce mapping of values so we have a valid starting value. This
        // constructor is used when reconstructing fields
        this->map(ptf, mapper);
    }
}


Foam::faradayConcentrationFvPatchScalarField::faradayConcentrationFvPatchScalarField
(
    const faradayConcentrationFvPatchScalarField& wbppsf
)
:
    fixedGradientFvPatchScalarField(wbppsf),
    molarMass_(wbppsf.molarMass_)
{}

Foam::faradayConcentrationFvPatchScalarField::faradayConcentrationFvPatchScalarField
(
    const faradayConcentrationFvPatchScalarField& wbppsf,
    const DimensionedField<scalar, volMesh>& iF
)
:
    fixedGradientFvPatchScalarField(wbppsf, iF),
    molarMass_(wbppsf.molarMass_)
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::faradayConcentrationFvPatchScalarField::updateCoeffs()
{

    if (updated())
    {
        return;
    }

    const double F = 96485.3329; // Faraday constant

    const dictionary& fluidDict = this->db().lookupObject<IOdictionary>
    (
        "transportProperties"
    );
    dimensionedScalar z("z", dimless, fluidDict);
    dimensionedScalar D("D", dimensionSet(0,2,-1,0,0,0,0), fluidDict);
    // get diffusion coefficient and number of electrons

    // current density
    const fvPatchField<vector>& J = patch().lookupPatchField<volVectorField, scalar>("J");

    // normal vector points out of the domain for 1-sided boundary
    // normal vector points from owner to neighbour cell for 2-sided boundary

    // gradient() = (-J / (z.value()*F*D.value()) * molarMass_) & patch().nf(); //in kg/m⁴
    gradient() = (-J / (z.value()*F*D.value()) ) & patch().nf(); //in mol/m⁴

    fixedGradientFvPatchScalarField::updateCoeffs();
}

void Foam::faradayConcentrationFvPatchScalarField::write(Ostream& os) const
{
    fixedGradientFvPatchScalarField::write(os);
    os.writeEntry("molarMass", molarMass_);
    writeEntry("value", os);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
    makePatchTypeField
    (
        fvPatchScalarField,
        faradayConcentrationFvPatchScalarField
    );
}
