#include "cyclicFvsPatchField.H"

// mappes surfaceScalarField back to global mesh
void Foam::mapMesh::rmapPatch
(
    const surfaceScalarField& fromF, // from Field
    surfaceScalarField& toF
)
{
    const fvMesh& fromMesh = fromF.mesh();
    const fvMesh& toMesh   = toF.mesh();

    //toF.correctBoundaryConditions(); // compute processor patches

    forAll (patchesMap_, patchI)
    {
        // _to_ patches (-1) and processor-patches do not need to be mapped
        if(patchesMap_[patchI] > -1 && patchesMap_[patchI] < toMesh.boundary().size())
        {
            // corresponding face numbers of parent mesh
            labelField curFpm // "current field patch mapper"
            (
                labelField::subField // a subfield of the facemap, of size of the patch
                (
                    faceMap_,
                    fromMesh.boundary()[patchI].size(),
                    fromMesh.boundary()[patchI].patch().start()
                )
            );
            // the above face numbers belong to the parent mesh; numbering is _global_
            // to get numbering related to the patch, subtract start of patch
            curFpm -= toMesh.boundary()[patchesMap_[patchI]].patch().start();

            toF.boundaryFieldRef()[patchesMap_[patchI]].
            scalarField::rmap(fromF.boundaryField()[patchI],curFpm);
        }
        else if (patchesMap_[patchI] == -1)  // _to_ patches
        // it seems, that _to_ patches contain the real face value, and not the neighbour cell value
        {
            labelField curFpm   // corresponding face number at parent mesh
            (
                labelField::subField
                (
                    faceMap_,
                    fromMesh.boundary()[patchI].size(),
                    fromMesh.boundary()[patchI].patch().start()
                )
            );

            // 1 if faces have same orientation; -1 if opposite orientation
            scalarField curMask // "current mask"
            (
                scalarField::subField
                (
                    faceMask_,
                    fromMesh.boundary()[patchI].size(),
                    fromMesh.boundary()[patchI].patch().start()
                )
            );

            // map directly to surfaceField, because parent mesh has no _to_ patches
            forAll (fromF.boundaryField()[patchI], faceI)
            {
                label curFace = curFpm[faceI];

                if(toMesh.isInternalFace(curFace))    // is an internal field in parent mesh
                {
                    toF[curFpm[faceI]] = fromF.boundaryField()[patchI][faceI]*curMask[faceI];
                }
                else
                {
                    label patchID = toMesh.boundaryMesh().whichPatch(curFace);
                    word patchName = toMesh.boundaryMesh()[patchID].name();
                    label start = toMesh.boundaryMesh()[patchID].start();
                    label localFaceID = curFace - start;
                                
                    toF.boundaryFieldRef()[patchID][localFaceID] = fromF.boundaryField()[patchI][faceI];
                    
                    if(isA<cyclicFvsPatchField<scalar>>(toF.boundaryField()[patchID]))
	                {  
                        const fvPatch& pID(toF.boundaryField()[patchID].patch());
                        const polyPatch& ppID(pID.patch());
                        
                        //make a mappedPatchBase from the surface patch
                        IOdictionary mpp0Dict
                        (
                            IOobject
                            (
                                "constant/mapPatchDict",
                                toMesh,
                                IOobject::MUST_READ,
                                IOobject::NO_WRITE
                            )
                        );

                        const dictionary& mppDict = mpp0Dict.subDict(patchName);

                        const mappedPatchBase mppID(ppID, mppDict);

                        // Get the coupling information from the mappedPatchBase
                        const mappedPatchBase& mpp =
                            refCast<const mappedPatchBase>(mppID);
                        
                        //get mesh to which the neighbour patch belongs
                        const polyMesh& nbrMesh = mpp.sampleMesh();
                        //get information about the neighbour patch
                        const label samplePatchI = mpp.samplePolyPatch().index();
                        const fvPatch& nbrPatch =
                            refCast<const fvMesh>(nbrMesh).boundary()[samplePatchI];
                      
                        toF.boundaryFieldRef()[samplePatchI][localFaceID] = -toF.boundaryField()[patchID][localFaceID];                       
                    }
                }
            }
        }      
    }
}


// mappes scalarField back to global mesh
void Foam::mapMesh::rmapPatch
(
    const volScalarField& fromF, // from Field
    volScalarField& toF
)
{
    const fvMesh& fromMesh = fromF.mesh();
    const fvMesh& toMesh   = toF.mesh();

    toF.correctBoundaryConditions(); // compute processor patches

    forAll (patchesMap_, patchI)
    {
	// _to_ patches (-1) and processor-patches do not need to be mapped
    	if(patchesMap_[patchI] > -1 && patchesMap_[patchI] < toMesh.boundary().size())
    	{
	    // corresponding face numbers of parent mesh
      	    labelField curFpm // "current field patch mapper"
      	    (
       	        labelField::subField // a subfield of the facemap, of size of the patch
       	        (
		    faceMap_,
		    fromMesh.boundary()[patchI].size(),
		    fromMesh.boundary()[patchI].patch().start()
       	        )
      	    );
            // the above face numbers belong to the parent mesh; numbering is _global_
	    // to get numbering related to the patch, subtract start of patch
      	    curFpm -= toMesh.boundary()[patchesMap_[patchI]].patch().start();

      	    toF.boundaryFieldRef()[patchesMap_[patchI]].
	        scalarField::rmap(fromF.boundaryField()[patchI],curFpm);
    	}        
    }
}


// mappes vectorField back to global mesh
void Foam::mapMesh::rmapPatch
(
    const volVectorField& fromF, // from Field
    volVectorField& toF
)
{
    const fvMesh& fromMesh = fromF.mesh();
    const fvMesh& toMesh   = toF.mesh();

    toF.correctBoundaryConditions(); // compute processor patches

    forAll (patchesMap_, patchI)
    {
        // _to_ patches (-1) and processor-patches do not need to be mapped
        if(patchesMap_[patchI] > -1 && patchesMap_[patchI] < toMesh.boundary().size())
        {
            // corresponding face numbers of parent mesh
            labelField curFpm // "current field patch mapper"
            (
                labelField::subField // a subfield of the facemap, of size of the patch
                (
                    faceMap_,
                    fromMesh.boundary()[patchI].size(),
                    fromMesh.boundary()[patchI].patch().start()
                )
            );
            // the above face numbers belong to the parent mesh; numbering is _global_
            // to get numbering related to the patch, subtract start of patch
            curFpm -= toMesh.boundary()[patchesMap_[patchI]].patch().start();

            // 1 if faces have same orientation; -1 if opposite orientation
            scalarField curMask // "current mask"
            (
                scalarField::subField
                (
                    faceMask_,
                    fromMesh.boundary()[patchI].size(),
                    fromMesh.boundary()[patchI].patch().start()
                )
            );

            toF.boundaryFieldRef()[patchesMap_[patchI]].
            vectorField::rmap(fromF.boundaryField()[patchI]*curMask,curFpm);
        }
    }
}


// maps patches of scalar field from global to child mesh; with weighting factor
void Foam::mapMesh::mapPatch
(
    const volScalarField& fromF, // fromField
    volScalarField& toF,
    const word& sigma // weighting factor
)
{
    const fvMesh& fromMesh = fromF.mesh();
    const fvMesh& toMesh   = toF.mesh();

    // field is weigthed by "sigma"
    std::stringstream ss;
    ss << "volWeighted " << "\"" << sigma << "\"";
    IStringStream scheme(ss.str());

    surfaceScalarField fromS = fvc::interpolate(fromF, scheme); // fromSurface

    toF.correctBoundaryConditions();      // compute processor patches

    forAll(toF.boundaryField(), patchI)
    {
	// here not: _to_ patch (-1) and processor patches
    	if(patchesMap_[patchI] > -1 && patchesMap_[patchI] < fromMesh.boundary().size())
	{
    	    labelField curFpm
    	    (
     	    	labelField::subField // subfield of faceMap_ of size of boundary to map for
     	    	(
      		    faceMap_,
      		    toMesh.boundary()[patchI].size(),
      		    toMesh.boundary()[patchI].patch().start()
     	    	)
    	    );
    	    curFpm -= fromMesh.boundary()[patchesMap_[patchI]].patch().start();

    	    forAll (toF.boundaryField()[patchI], faceI)
    	    {
      	    	toF.boundaryFieldRef()[patchI][faceI] = 
		  fromF.boundaryField()[patchesMap_[patchI]][curFpm[faceI]];
    	    }
    	}
        else if (patchesMap_[patchI] == -1)  // _to_ patches
        // it seems, that _to_ patches contain the real face value, and not the neighbour cell value
        {
            labelField curFpm   // corresponding face number at parent mesh
            (
                labelField::subField
                (
                    faceMap_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );

            // map directly from surfaceField, because parent mesh has no _to_ patches
            forAll (toF.boundaryField()[patchI], faceI)
            {
                label curFace = curFpm[faceI];

                if(fromMesh.isInternalFace(curFace))    // is an internal field in parent mesh
                {
                    toF.boundaryFieldRef()[patchI][faceI] = fromS[curFpm[faceI]];
                }
                else					// is a processor or external face
                {
                    label patchID = fromMesh.boundaryMesh().whichPatch(curFace);
                    label start = fromMesh.boundaryMesh()[patchID].start();
                    label localFaceID = curFace - start;

                    toF.boundaryFieldRef()[patchI][faceI] = fromS.boundaryField()[patchID][localFaceID];
                }
            }
        }
    }
}


// maps patches of vectorField from global to child mesh; with weighting factor
void Foam::mapMesh::mapPatch
(
    const volVectorField& fromF, // fromField
    volVectorField& toF,
    const word& sigma // weighting factor
)
{
    const fvMesh& fromMesh = fromF.mesh();
    const fvMesh& toMesh   = toF.mesh();

    // field is weigthed by "sigma"
    std::stringstream ss;
    ss << "volWeighted " << "\"" << sigma << "\"";
    IStringStream scheme(ss.str());

    surfaceVectorField fromS = fvc::interpolate(fromF, scheme); // fromSurface

    toF.correctBoundaryConditions();      // compute processor patches

    forAll(toF.boundaryField(), patchI)
    {
        // here not: _to_ patch (-1) and processor patches
        if(patchesMap_[patchI] > -1 && patchesMap_[patchI] < fromMesh.boundary().size())
        {
            labelField curFpm
            (
                labelField::subField // subfield of faceMap_ of size of boundary to map for
                (
                    faceMap_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );
            curFpm -= fromMesh.boundary()[patchesMap_[patchI]].patch().start();

            scalarField curMask
            (
                scalarField::subField
                (
                    faceMask_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );

            forAll (toF.boundaryField()[patchI], faceI)
            {
                toF.boundaryFieldRef()[patchI][faceI] =
                  fromF.boundaryField()[patchesMap_[patchI]][curFpm[faceI]]*curMask[faceI];
            }
        }
        else if (patchesMap_[patchI] == -1)  // _to_ patches
        // it seems that _to_ patches contain the real face value, and not the neighbour cell value
        {
            labelField curFpm   // corresponding face number at parent mesh
            (
                labelField::subField
                (
                    faceMap_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );

            scalarField curMask
            (
                scalarField::subField
                (
                    faceMask_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );

            // map directly from surfaceField, because parent mesh has no _to_ patches
            forAll (toF.boundaryField()[patchI], faceI)
            {
                label curFace = curFpm[faceI];

                if(fromMesh.isInternalFace(curFace))    // is an internal field in parent mesh
                {
                    toF.boundaryFieldRef()[patchI][faceI] = fromS[curFpm[faceI]]*curMask[faceI];
                }
                else                                    // is a processor or external face
                {
                    label patchID = fromMesh.boundaryMesh().whichPatch(curFace);
                    label start = fromMesh.boundaryMesh()[patchID].start();
                    label localFaceID = curFace - start;
		    
		    // here is no mask - is that correct?
                    toF.boundaryFieldRef()[patchI][faceI] = fromS.boundaryField()[patchID][localFaceID];
                }
            }
        }
    }
}


// maps patches of scalarField from global to child mesh, no weighting
void Foam::mapMesh::mapPatch
(
    const volScalarField& fromF, // fromField
    volScalarField& toF
)
{
    const fvMesh& fromMesh = fromF.mesh();
    const fvMesh& toMesh   = toF.mesh();
    // _to_ patches are mapped using a surfaceScalarField
    surfaceScalarField fromS = linearInterpolate(fromF); // fromSurface

    toF.correctBoundaryConditions();    // compute processor patches
    forAll(toF.boundaryField(), patchI)
    {
  	if(patchesMap_[patchI] > -1 && patchesMap_[patchI] < fromMesh.boundary().size())
  	{
    	    labelField curFpm
    	    (
     	    	labelField::subField
     	    	(
      		    faceMap_,
      		    toMesh.boundary()[patchI].size(),
      		    toMesh.boundary()[patchI].patch().start()
     		)
    	    );
    	    curFpm -= fromMesh.boundary()[patchesMap_[patchI]].patch().start();

    	    forAll (toF.boundaryField()[patchI], faceI)
    	    {
      		toF.boundaryFieldRef()[patchI][faceI] = 
		  fromF.boundaryField()[patchesMap_[patchI]][curFpm[faceI]];
    	    }
  	}
  	else if (patchesMap_[patchI] == -1)  // _to_ patches
	// it seems, that _to_ patches contain the real face value, and not the neighbour cell value
	{
            labelField curFpm	// corresponding face number at parent mesh
            (
                labelField::subField
                (
                    faceMap_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );

	    // map directly from surfaceField, because parent mesh has no _to_ patches
            forAll (toF.boundaryField()[patchI], faceI)
            {
		label curFace = curFpm[faceI];

		if(fromMesh.isInternalFace(curFace))	// is an internal field in parent mesh
		{
		    toF.boundaryFieldRef()[patchI][faceI] = fromS[curFpm[faceI]];
		}
		else
		{
		    label patchID = fromMesh.boundaryMesh().whichPatch(curFace);
		    label start = fromMesh.boundaryMesh()[patchID].start();
		    label localFaceID = curFace - start;

		    toF.boundaryFieldRef()[patchI][faceI] = fromS.boundaryField()[patchID][localFaceID];
		}
            }
  	}
    }
}


// maps patches of vectorField from global to child mesh, no weighting
void Foam::mapMesh::mapPatch
(
    const volVectorField& fromF, // fromField
    volVectorField& toF
)
{
    const fvMesh& fromMesh = fromF.mesh();
    const fvMesh& toMesh   = toF.mesh();
    // _to_ patches are mapped using a surfaceScalarField

    surfaceVectorField fromS = linearInterpolate(fromF); // fromSurface

    toF.correctBoundaryConditions();    // compute processor patches
    forAll(toF.boundaryField(), patchI)
    {
        if(patchesMap_[patchI] > -1 && patchesMap_[patchI] < fromMesh.boundary().size())
        {
            labelField curFpm
            (
                labelField::subField
                (
                    faceMap_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );
            curFpm -= fromMesh.boundary()[patchesMap_[patchI]].patch().start();

            scalarField curMask
            (
                scalarField::subField
                (
                    faceMask_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );

            forAll (toF.boundaryField()[patchI], faceI)
            {
                toF.boundaryFieldRef()[patchI][faceI] =
                  fromF.boundaryField()[patchesMap_[patchI]][curFpm[faceI]]*curMask[faceI];
            }
        }
        else if (patchesMap_[patchI] == -1)  // _to_ patches
        // it seems, that _to_ patches contain the real face value, and not the neighbour cell value
        {
            labelField curFpm   // corresponding face number at parent mesh
            (
                labelField::subField
                (
                    faceMap_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );

            scalarField curMask
            (
                scalarField::subField
                (
                    faceMask_,
                    toMesh.boundary()[patchI].size(),
                    toMesh.boundary()[patchI].patch().start()
                )
            );

            // map directly from surfaceField, because parent mesh has no _to_ patches
            forAll (toF.boundaryField()[patchI], faceI)
            {
                label curFace = curFpm[faceI];

                if(fromMesh.isInternalFace(curFace))    // is an internal field in parent mesh
                {
                    toF.boundaryFieldRef()[patchI][faceI] = fromS[curFpm[faceI]];//*curMask[faceI];
                }
                else
                {
                    label patchID = fromMesh.boundaryMesh().whichPatch(curFace);
                    label start = fromMesh.boundaryMesh()[patchID].start();
                    label localFaceID = curFace - start;

		    // here without mask - is that correct?
                    toF.boundaryFieldRef()[patchI][faceI] = fromS.boundaryField()[patchID][localFaceID];
                }
            }
        }
    }
}
