/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Description
    Similar as a cell zone: a subset of mesh cells, but with a number 
    of further properties.

Developed by
    Carolina Duczek and Norbert Weber

\*---------------------------------------------------------------------------*/

#ifndef solidZone_H
#define solidZone_H

#include "fvMesh.H"
#include "dictionary.H"
#include "fvMatricesFwd.H"
#include "runTimeSelectionTables.H"
#include "coordinateSystem.H"
#include "dimensionedVector.H"
#include "keyType.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{

/*---------------------------------------------------------------------------*\
                        Class solidZone Declaration
\*---------------------------------------------------------------------------*/

class solidZone
:
    public regIOobject
{
    // Private Member Functions

        //- Disallow default bitwise copy construct
        solidZone(const solidZone&);

        //- Disallow default bitwise assignment
        void operator=(const solidZone&);


protected:

    // Protected data

        //- Reference to the mesh database
        const fvMesh& mesh_;

        //- Dictionary used for model construction
        const dictionary dict_;

        //- Porosity active flag
        bool active_;

        //- Name(s) of cell-zone
        keyType zoneName_;

        //- Cell zone IDs
        labelList cellZoneIDs_;

        dimensionedScalar k_;
	dimensionedScalar cp_;
	dimensionedScalar rho_;
        dimensionedScalar sigma_;

public:

    //- Runtime type information
    TypeName("solidZone");

    //- Selection table
    declareRunTimeSelectionTable
    (
        autoPtr,
        solidZone,
        mesh,
        (
            const fvMesh& mesh,
            const dictionary& dict,
            const word& cellZoneName
        ),
        (mesh, dict, cellZoneName)
    );

    //- Constructor
    solidZone
    (
        const fvMesh& mesh,
        const dictionary& dict,
        const word& cellZoneName = word::null
    );

    //- Return pointer to new solidZone object created on the freestore
    //  from an Istream
    class iNew
    {
        //- Reference to the mesh database
        const fvMesh& mesh_;

    public:

        iNew
        (
            const fvMesh& mesh
        )
        :
            mesh_(mesh)
        {}

        autoPtr<solidZone> operator()(Istream& is) const
        {
            const dictionary dict(is);

            return autoPtr<solidZone>
            (
                solidZone::New
                (
                    mesh_,
                    dict
                )
            );
        }
    };

    //- Selector
    static autoPtr<solidZone> New
    (
        const fvMesh& mesh,
        const dictionary& dict,
        const word& cellZoneName = word::null
    );

    //- Destructor
    virtual ~solidZone();


    // Member Functions

        //- Return const access to the zone name

        inline const keyType& zoneName() const;

	//- dictionary values used for the solidZone

        inline const dictionary& dict() const;

        //- Return const access to the porosity active flag
        inline bool active() const;

        //- Return const access to the cell zone IDs
        inline const labelList& zoneIds() const;

        dimensionedScalar k() const;
        dimensionedScalar cp() const;
        dimensionedScalar rho() const;
 	dimensionedScalar sigma() const;

    // I-O

        //- Write
        virtual bool writeData(Ostream& os) const;

        //- Inherit read from regIOobject
        using regIOobject::read;

        //- Read porosity dictionary
        virtual bool read(const dictionary& dict);
};


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace Foam

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#include "solidZoneI.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

#endif

// ************************************************************************* //
