/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "mapMesh.H"
#include "mapPatch.H"


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::mapMesh::mapMesh
(
    const fvMesh& mesh
)
:
    mesh_(mesh),
    faceRegionAddressing_
    (
        IOobject
        (
            "faceRegionAddressing",
            mesh.time().findInstance(mesh.meshDir(), "faces"),
            polyMesh::meshSubDir,
            mesh,
            IOobject::MUST_READ
        )
    ),
    faceMap_(faceRegionAddressing_.size()),
    faceMask_(faceRegionAddressing_.size()),
    cellMap_
    (
        IOobject
        (
            "cellRegionAddressing",
            mesh.time().findInstance(mesh.meshDir(), "faces"),
            polyMesh::meshSubDir,
            mesh,
            IOobject::MUST_READ
        )
    ),
    patchesMap_
    (
        IOobject
        (
            "boundaryRegionAddressing",
            mesh.time().findInstance(mesh.meshDir(), "faces"),
            polyMesh::meshSubDir,
            mesh,
            IOobject::MUST_READ
        )
    )
{
    // faceMask: same sign - same orientation of faces
    // faceMask: different sign - opposite orientation of faces
    // as "0" cannot have a sign, numbering starts with "1"
    // therefore, 1 needs to be subtracted from faceRegionA. to get the mapping

    forAll(faceMap_, i)
    {
        faceMap_[i] = mag(faceRegionAddressing_[i]) - 1;
        faceMask_[i] = sign(faceRegionAddressing_[i]);
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

void Foam::mapMesh::rmap
(
    volScalarField& targetField,
    const volScalarField& sourceField
)
{
    scalarField& targetFieldIn = targetField.primitiveFieldRef();
    targetFieldIn.rmap(sourceField, cellMap_);
    rmapPatch(sourceField, targetField);
}

void Foam::mapMesh::rmap
(
    volVectorField& targetField,
    const volVectorField& sourceField
)
{
    vectorField& targetFieldIn = targetField.primitiveFieldRef();
    targetFieldIn.rmap(sourceField, cellMap_);
    rmapPatch(sourceField, targetField);
}


void Foam::mapMesh::rmap
(
    volScalarField& targetField,
    const dimensionedScalar& sourceValue
)
{
    scalarField& targetFieldIn = targetField.primitiveFieldRef();
    scalarField sourceField(mesh_.nCells(), sourceValue.value());
    targetFieldIn.rmap(sourceField, cellMap_);
    //rmapPatch(sourceField, targetField); // todo
}


void Foam::mapMesh::rmap
(
    volScalarField& targetField,
    scalarField& sourceField
)
{
    scalarField& targetFieldIn = targetField.primitiveFieldRef();
    targetFieldIn.rmap(sourceField, cellMap_);
}


void Foam::mapMesh::rmap
(
    surfaceScalarField& targetField,
    const surfaceScalarField& sourceField
)
{
    // internal faces
    labelList internalFaceMap
    (
        SubList<label>(faceMap_, mesh_.nInternalFaces())
    );

    scalarField internalFaceMask
    (
        scalarField::subField(faceMask_, mesh_.nInternalFaces())
    );

    // map internal faces first
    // mask is 1 or -1 ... depending whether faces have same direction or not
    scalarField& targetFieldIn = targetField.primitiveFieldRef();
    targetFieldIn.rmap(sourceField.primitiveField()*internalFaceMask, internalFaceMap);

    rmapPatch(sourceField, targetField);
}


void Foam::mapMesh::map
(
    const volVectorField& sourceField,
    volVectorField& targetField
)
{
    forAll (targetField, cellI)
    {
        targetField[cellI] = sourceField[cellMap_[cellI]];
    }
    mapPatch(sourceField, targetField);
}

void Foam::mapMesh::map
(
    const volScalarField& sourceField,
    volScalarField& targetField
)
{
    forAll (targetField, cellI)
    {
        targetField[cellI] = sourceField[cellMap_[cellI]];
    }
    mapPatch(sourceField, targetField);
}

// ************************************************************************* //
