/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "speciesConcentrationFvPatchScalarField.H"
#include "fvPatchFieldMapper.H"
#include "volFields.H"
#include "surfaceFields.H"
#include "addToRunTimeSelectionTable.H"
#include "specie.H"

// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

Foam::speciesConcentrationFvPatchScalarField::speciesConcentrationFvPatchScalarField
(
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF
)
:
    fixedGradientFvPatchScalarField(p, iF),
    molarMass_(1.0),
    concentration_(this->size(), Zero) 
{}

// this constructor is called in the solver and decomposePar
Foam::speciesConcentrationFvPatchScalarField::speciesConcentrationFvPatchScalarField
(
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF,
    const dictionary& dict
)
:
    fixedGradientFvPatchScalarField(p, iF),
    molarMass_(dict.lookupOrDefault<scalar>("molarMass", 1.0)),
    concentration_(p.size(), Zero)
{
    concentration_ = Field<scalar>("value", dict, p.size());
    
    if (dict.found("value"))                    // initial value
    {
        fvPatchField<scalar>::operator=
        (
            scalarField("value", dict, p.size())
        );
    }
    else
    {
        fvPatchField<scalar>::operator=(patchInternalField());
    }
}

// probably only needed for mapFields
Foam::speciesConcentrationFvPatchScalarField::speciesConcentrationFvPatchScalarField
(
    const speciesConcentrationFvPatchScalarField& ptf,
    const fvPatch& p,
    const DimensionedField<scalar, volMesh>& iF,
    const fvPatchFieldMapper& mapper
)
:
    fixedGradientFvPatchScalarField(p, iF),
    molarMass_(ptf.molarMass_),
    concentration_(ptf.concentration_)
{
    patchType() = ptf.patchType();

    // Map gradient. Set unmapped values and overwrite with mapped ptf
    gradient() = 0.0;
    gradient().map(ptf.gradient(), mapper);

    // Evaluate the value field from the gradient if the internal field is valid
    if (notNull(iF) && iF.size())
    {
        scalarField::operator=
        (
            //patchInternalField() + gradient()/patch().deltaCoeffs()
            // ***HGW Hack to avoid the construction of mesh.deltaCoeffs
            // which fails for AMI patches for some mapping operations
            patchInternalField() + gradient()*(patch().nf() & patch().delta())
        );
    }
    else
    {
        // Enforce mapping of values so we have a valid starting value. This
        // constructor is used when reconstructing fields
        this->map(ptf, mapper);
    }
}
    
Foam::speciesConcentrationFvPatchScalarField::speciesConcentrationFvPatchScalarField
(
    const speciesConcentrationFvPatchScalarField& wbppsf
)
:
    fixedGradientFvPatchScalarField(wbppsf),
    molarMass_(wbppsf.molarMass_),
    concentration_(wbppsf.concentration_)
{}

Foam::speciesConcentrationFvPatchScalarField::speciesConcentrationFvPatchScalarField
(
    const speciesConcentrationFvPatchScalarField& wbppsf,
    const DimensionedField<scalar, volMesh>& iF
)
:
    fixedGradientFvPatchScalarField(wbppsf, iF),
    molarMass_(wbppsf.molarMass_),
    concentration_(wbppsf.concentration_)
{}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

Foam::tmp<Foam::Field<scalar>> Foam::speciesConcentrationFvPatchScalarField::value() const
{
    return concentration_;
}

 
void Foam::speciesConcentrationFvPatchScalarField::updateCoeffs()
{

    if (updated())
    {
        return;
    }

    word fieldName = this->internalField().name().substr(this->internalField().name().find(".")+1,-1);
    
    const dimensionedScalar F("F", dimensionSet(0,0,1,0,-1,1,0), 96485.3329); // Faraday constant
    
    //current density
    const fvPatchField<vector>& J = patch().lookupPatchField<volVectorField, scalar>("J");
    
    //get species object which belongs to this concentration
    const specie& sp = this->db().lookupObject<specie>(this->internalField().name());
    const fvPatchField<scalar>& c = patch().lookupPatchField<volScalarField, scalar>("c."+sp.name());
    
    //needed to sum up nu^2*c
    scalarField nu2c=c*0;

    //get list of all species
    speciesMixture& sm = this->db().lookupObjectRef<speciesMixture>("speciesProperties");
    const PtrList<specie>& species = sm.species();

   //calculate an intermediate parameter (nu2c) which appears in concentration gradient expression
   forAll(species,is)
    {
        const dimensionedScalar nuis = species[is].nu();
        const fvPatchField<scalar>& cis = patch().lookupPatchField<volScalarField, scalar>("c."+species[is].name());
        nu2c  += nuis.value()*nuis.value()*cis;
    }

    //get species object which belongs to this concentration
    //const specie& sp = this->db().lookupObject<specie>(this->internalField().name());
    const dimensionedScalar nu = sp.nu();
    const dimensionedScalar D = sp.D();
    
    //look up if species is active or not
    //properties of the active species are defined in constant/salt/transportProperties
    const bool active = sp.active();  
    
    const dictionary& solidDict = this->db().lookupObject<IOdictionary>
    (
     "speciesProperties"
    );
         
    const dimensionedScalar DActive("DActive", dimensionSet(0,2,-1,0,0,0,0), solidDict);
    const dimensionedScalar nuActive("nuActive", dimensionSet(0,0,0,0,0,0,0), solidDict);
    
    //NActive = J / (F.value()*nuActive.value());
                     
    if(active)
    {
    	Info<< "species " << fieldName << " is active" << endl;
      	gradient() = J / (F.value()*nuActive.value()) / DActive.value() * (nu.value()*nu.value()*c/nu2c - 1) & patch().nf();
      	//gradient() = (J / (nu.value()*F.value()*DActive.value()) * (-1+nu.value()*nu.value()*c/nu2c)) & patch().nf(); //for a binary electrolyte
    }
    else
    {
    	Info<< "species " << fieldName << " is not active" << endl;
      	gradient() = J / (F.value()*nuActive.value()) / DActive.value() * nu.value()*nuActive.value()*c/nu2c & patch().nf();
      	//gradient() = (J*nu.value()*c / (F.value()*DActive.value()*nu2c)) & patch().nf(); //for a binary electrolyte
    }

    fixedGradientFvPatchScalarField::updateCoeffs();
}

void Foam::speciesConcentrationFvPatchScalarField::write(Ostream& os) const
{
    fixedGradientFvPatchScalarField::write(os);
    os.writeKeyword("molarMass") << molarMass_ << token::END_STATEMENT << nl;;
    writeEntry("value", os);
}

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
    makePatchTypeField
    (
        fvPatchScalarField,
        speciesConcentrationFvPatchScalarField
    );
}

// ************************************************************************* //
