// solves Poisson eqn. for the electric potential

Info<< "start calculating potential equation (poisson)..." << endl; 

// the brackets make sure that variables are only defined inside the brackets
// (avoids mistakes)
{
    if (finalIter)
    {
        globalMesh.data::add("finalIteration", true);
    }
 
    for (int corr=0; corr<nCorr; corr++)
    {
        // get a reference to the global fields
        // (only to use shorter names)
        volScalarField& sigma = sigmaGlobal;
        volScalarField& potential = potentialGlobal;
        volVectorField& gradPotential = gradPotentialGlobal;
        surfaceScalarField& gradPotentialSurface = gradPotentialSurfaceGlobal;
        volVectorField& j = jGlobal;
        volVectorField& jd = jdGlobal;
        volVectorField& jm = jmGlobal;
        volScalarField& sourceTerm = sourceTermGlobal;
    
        // potential in electrolyte has to be solved considering species concentration and transport due to migration and diffusion

        // electric conductivity in salt is dependent on the concentration of the species
        forAll(electrolyteRegions, i) 
        {   	
            volScalarField& sigmaSalt = sigmaEl[i];
    	    speciesMixture& speciesMix = speciesMixtureEl[i];
    	
    	    // set sigmaSalt to zero since it is a sum
    	    sigmaSalt = sigmaSalt * 0;
    	
    	    forAll(speciesMix.species(), is)
    	    {
    		volScalarField& c = speciesMix.species()[is];
    		const dimensionedScalar& nu = speciesMix.species()[is].nu();
    		const dimensionedScalar& D = speciesMix.species()[is].D();
    		const dimensionedScalar mu("mu", dimensionSet(-1,0,1,0,1,0,0), D.value()/R.value()/T0.value());
    		sigmaSalt += F*F*nu*nu*mu*c; 

	    }
	
	    Info << "calculated conductivity in salt" << endl;
	
	    // map the electric conductivity to the global mesh
    	    electrolyteMesh[i].rmap(sigmaGlobal, sigmaEl[i]);
        }
        sigma.correctBoundaryConditions(); 
    
        // solve potential equation
        // has to be solved only on the global mesh
    
        for (int nonOrth=0; nonOrth<=nNonOrthCorr; nonOrth++)
        {
            fvScalarMatrix potentialEqn
            (
                fvm::laplacian(sigma, potential)
            );
	
            // in the electrolyte the current density contains migration and diffusion of the species
            forAll(electrolyteRegions, i)
            {
                surfaceScalarField& jdSurfaceSalt = jdSurfaceEl[i];
                speciesMixture& speciesMix = speciesMixtureEl[i];
        
                electrolyteMesh[i].map(potentialGlobal, potentialEl[i]);
        
                // set jdSalt to zero since it is a sum
                jdSurfaceSalt = jdSurfaceSalt * 0;
    
                forAll(speciesMix.species(), is)
                {	
                    volScalarField& c = speciesMix.species()[is];
                    const dimensionedScalar& nu = speciesMix.species()[is].nu();
                    const dimensionedScalar& D = speciesMix.species()[is].D();
                    jdSurfaceSalt -= F*nu*D*fvc::snGrad(c);
                }
        
                // map the current density due to diffusion to the global mesh
                electrolyteMesh[i].rmap(jdSurfaceGlobal, jdSurfaceEl[i]);
            }

            // potential equation contains a source term for diffusion
            // the source term is nonzero only in the cells of the electrolyte
            sourceTerm = fvc::div(jdSurfaceGlobal*globalMesh.magSf()); // source term for electric current
      
            // the potential is determined solving a Poisson equation
            // we need the values on the cell faces for a consistent discretisation (balanced force)
            potentialEqn -= sourceTerm;
        
            potentialEqn.solve
            (
                globalMesh.solver
                (
                    potential.select
                    (
                        oCorr == nOuterCorr-1
                     && corr == nCorr-1
                     && nonOrth == nNonOrthCorr    
                    )
                )
            );
        }

        gradPotential = fvc::grad(potential);
        gradPotentialSurface = fvc::snGrad(potential);
        
        // get cell value from surface value after solving poisson equation on the faces
        jd = fvc::reconstruct(jdSurfaceGlobal * globalMesh.magSf());
    
        // reconstruct does not preserve the boundary values
        // it simply uses the cell centre value for the boundary, as well
        // therefore: write correct boundary values
        forAll(globalMesh.boundary(), b)
        {
	    jd.boundaryFieldRef()[b] = jdSurfaceGlobal.boundaryField()[b]*globalMesh.boundary()[b].nf();
        }
    
        // calculate current density considering migration and diffusion  
    
        jm = fvc::reconstruct
    	(
    	    -fvc::interpolate(sigma) * fvc::snGrad(potential) * globalMesh.magSf()
    	);
    	
    
        forAll(globalMesh.boundary(), b)
        {
            jm.boundaryFieldRef()[b] = jmSurfaceGlobal.boundaryField()[b]*globalMesh.boundary()[b].nf();
        }
       
        j = jm+jd;
    }
  
    if (finalIter)
    {
        globalMesh.data::remove("finalIteration");
    }
}

Info<< "stop calculating potential equation (poisson)" << endl; 
