// sets the boundary and initial conditions for current and concentration

// (1) electric current expressed as potential gradient

// read current from file
const dimensionedScalar I("I", dimCurrent, globalProperties);

// find patch "top"
label topPatch = globalMesh.boundaryMesh().findPatchID("top");

// get area of the anode
dimensionedScalar Atop("0", dimensionSet(0,2,0,0,0,0,0), Foam::gSum(globalMesh.boundary()[topPatch].magSf()));
Info<< "Atop is " << Atop.value() << " m²" << endl;

// calculate current density 
const dimensionedScalar j0top = I / Atop;
Info<< "j0top is " << j0top.value() << " V/m" << endl;
Info<< "Adjusting electric current to " << I.value() << " A at top boundary" << endl;

// set Neumann BC for Potential
if(isA<fixedGradientFvPatchScalarField>(potentialGlobal.boundaryField()[topPatch]))
{
    fixedGradientFvPatchField<scalar>& potentialTop = 
        static_cast<fixedGradientFvPatchField<scalar>&>(potentialGlobal.boundaryFieldRef()[topPatch]);
    if(potentialTop.size() > 0)
    {
	potentialTop.gradient() = -j0top.value()/sigmaGlobal.boundaryField()[topPatch];
    }
} 

Info<< "sigmaTop is " << sigmaGlobal.boundaryField()[topPatch]*1 << " S/m" << endl;
Info<< "Adjusting potential gradient to " << -j0top.value()/sigmaGlobal.boundaryField()[topPatch] << " V at top boundary" << endl;


// (2) sets the initial condition for the concentration
Info<< "rho0Ges = " << rho0Ges.value() << " kg/m^3" << endl;

const dimensionedScalar c0 = x0 * rho0Ges / (x0*MLi+(1-x0)*MBi);

Info<< "Adjusting initial concentration of Li in Bi to " << c0.value() << " mol/m^3" << endl;

forAll(fluidRegions, i)
{
    if(fluidRegions[i].name() == "Bi")
    {
    	volScalarField& c = cFluids[i];
    
    	forAll(c, b)
    	{
    	    double& cInitial = static_cast<double&>(c.ref()[b]);  
    	    cInitial = c0.value();
    	}    
    }
}

forAll(electrolyteRegions, i)
{
    speciesMixture& speciesMix = speciesMixtureEl[i];
	
    forAll(speciesMix.species(), is)
    {
    	// find patch for IF salt-anode and salt-cathode
	label speciesAnodePatch = electrolyteRegions[i].boundaryMesh().findPatchID("salt_to_Li");
	label speciesCathodePatch = electrolyteRegions[i].boundaryMesh().findPatchID("salt_to_Bi");
    		
 	volScalarField& cSpecies = speciesMix.species()[is];
    		
  	// read concentration at patch so that it can be used for the jump condition
	if(isA<speciesConcentrationFvPatchScalarField>(cSpecies.boundaryField()[speciesAnodePatch]))
	{
	    speciesConcentrationFvPatchScalarField& speciesConcentrationAnode = 
                static_cast<speciesConcentrationFvPatchScalarField&>(cSpecies.boundaryFieldRef()[speciesAnodePatch]);
	    if(speciesConcentrationAnode.size() > 0)
	    {
		Info<< "concentration of species " << cSpecies.name() 
                    << " at IF anode-salt is " << speciesConcentrationAnode.value() << " mol/m^3" << endl;
	    }
	}
		
	if(isA<speciesConcentrationFvPatchScalarField>(cSpecies.boundaryField()[speciesCathodePatch]))
	{
	    speciesConcentrationFvPatchScalarField& speciesConcentrationCathode =  
                static_cast<speciesConcentrationFvPatchScalarField&>(cSpecies.boundaryFieldRef()[speciesCathodePatch]);
	    if(speciesConcentrationCathode.size() > 0)
	    {
		Info<< "concentration of species " << cSpecies.name() 
                    << " at IF cathode-salt is " << speciesConcentrationCathode.value() << " mol/m^3" << endl;
	    }
	}        
    }
}


// (3) calculate electric conductivity in the salt
// electric conductivity in salt is dependent on the concentration of the species
forAll(electrolyteRegions, i)
{   	
    volScalarField& sigmaSalt = sigmaEl[i];
    speciesMixture& speciesMix = speciesMixtureEl[i];
    	
    // set sigmaSalt to zero since it is a sum
    sigmaSalt = sigmaSalt * 0;
    	
    forAll(speciesMix.species(), is)
    {
    	volScalarField& c = speciesMix.species()[is];
    	const dimensionedScalar& nu = speciesMix.species()[is].nu();
    	const dimensionedScalar& D = speciesMix.species()[is].D();
    	const dimensionedScalar mu("mu", dimensionSet(-1,0,1,0,1,0,0), D.value()/R.value()/T0.value());
    	sigmaSalt += F*F*nu*nu*mu*c; 
    }

    // map the electric conductivity to the global mesh
    electrolyteMesh[i].rmap(sigmaGlobal, sigmaEl[i]);
}

Info<< "Adjusting electric conductivity on the global mesh " << endl;


// (4) calculate initial current density on the global mesh
jGlobal = -sigmaGlobal * fvc::grad(potentialGlobal);

Info<< "set initial current density on the global mesh" << endl;

forAll(electrolyteRegions, i)
{
    volVectorField& jdSalt = jdEl[i];
    speciesMixture& speciesMix = speciesMixtureEl[i];
    	
    // set jdSalt to zero since it is a sum
    jdSalt = jdSalt * 0;
    	
    forAll(speciesMix.species(), is)
    {	
    	volScalarField& c = speciesMix.species()[is];
    	const dimensionedScalar& nu = speciesMix.species()[is].nu();
    	const dimensionedScalar& D = speciesMix.species()[is].D();
    	jdSalt -= F*nu*D*fvc::grad(c);   
    }
    	
    // map the current density du to migration to the global mesh
    electrolyteMesh[i].rmap(jdGlobal, jdEl[i]);
}
// add diffusive current 
jGlobal += jdGlobal;

Info<< "added diffusion term to current density on the global mesh" << endl;
