/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     |
    \\  /    A nd           | www.openfoam.com
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

Application
    speciesFoam

Description
    Solver for batteries with molten electrodes and molten salt electrolyte. 
    Species equations are solved in the electrolyte, and the electric
    potential and current in the complete geometry

Publication
    https://doi.org/10.1016/j.electacta.2022.141413
    https://arxiv.org/abs/2211.12466

Developed by
    Carolina Duczek

\*---------------------------------------------------------------------------*/

#include "fvCFD.H"
#include "regionProperties.H"
#include "solidZoneList.H"
#include "mapMesh.H"
#include "potentialJumpFvPatchField.H"
#include "faradayConcentrationFvPatchScalarField.H"
#include "speciesConcentrationFvPatchScalarField.H"
#include "fvPatchField.H"
#include "fvMatrix.H"
#include "speciesMixture.H"
#include "specie.H"
#include "fvOptions.H"
#include "pimpleControl.H"
#include "harmonic.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

int main(int argc, char *argv[])
{
    #include "setRootCaseLists.H"

    #include "createTime.H"

    #include "readPIMPLEControlsResiduals.H"
    #include "createGlobalMesh.H"
    #include "createFluidMeshes.H"
    #include "createElectrolyteMeshes.H"  

    #include "createGlobalFields.H"
    #include "createFluidFields.H"
    #include "createElectrolyteFields.H"  
    
    // * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //
    
    // map conductivity to global mesh
    #include "mapSigma.H"
    // load material parameter
    #include "material.H"
    // solve initial timestep
    #include "initialSolve.H"
       
    Info<< "\nStarting time loop\n" << endl;
    
    while (runTime.run())
    {   	
    	runTime++;     	
    	Info<< "Time = " << runTime.timeName() << nl << endl;
       
    	for (int oCorr=0; oCorr<nOuterCorr; oCorr++)
    	{
    	    Info<< "Pimple iteration " << oCorr << "\n";
            	
            bool finalIter = oCorr == nOuterCorr-1;
            	
            // solve species transport in salt
            #include "speciesEqn.H"
        	
            // solve electric potential and current with source term
            #include "potentialEqn_poisson.H"

            // solve concentration of Li in Bi (cathode)
            #include "cEqn.H"     
    	}

    	runTime.write();
    	
    	forAll(fluidRegions, i)
    	{
    	    if(fluidRegions[i].name() == "Bi")
    	    {
		runTime.write();
	    }
		
	    runTime.write();
    	}
    	
    	forAll(electrolyteRegions, i)
    	{  		
	    runTime.write();
		
	    forAll(speciesMixtureEl[i].species(), is)
    	    {
    		speciesMixtureEl[i].species()[is].write(); // writes concentration for species 
    	    }	
    	}
    	
    	Info<< "ExecutionTime = " << runTime.elapsedCpuTime() << " s"
            << "  ClockTime = " << runTime.elapsedClockTime() << " s"
            << nl << endl;
    }
    
    Info<< "End\n" << endl;
    
    return 0;
}

// ************************************************************************* //
